#![no_main]
#![no_std]

extern crate alloc;

use alloc::vec::Vec;

use core::iter::repeat;

use uefi::{
    prelude::*,
    proto::console::gop::{BltOp::BufferToVideo, BltPixel, BltRegion, GraphicsOutput},
    Result,
};

const ALIVE: BltPixel = BltPixel::new(0xff, 0xff, 0xff);
const DEAD: BltPixel = BltPixel::new(0, 0, 0);

#[derive(Debug, Clone)]
struct Life {
    height: usize,
    matrix: Vec<BltPixel>,
    width: usize,
}

impl Life {
    fn new(width: usize, height: usize) -> Self {
        let mut matrix = Vec::default();
        for _ in 0..height {
            matrix.extend(repeat(DEAD).take(width));
        }
        Life {
            matrix,
            width,
            height,
        }
    }

    fn with_starter(width: usize, height: usize, starter: &[(usize, usize)]) -> Self {
        let mut life = Self::new(width, height);
        for (i, j) in starter.iter() {
            life.matrix[i * width + j] = ALIVE;
        }
        life
    }

    fn play(self) -> Life {
        const OFFSETS: &[(i64, i64)] = &[
            (-1, -1),
            (-1, 0),
            (-1, 1),
            (0, -1),
            (0, 1),
            (1, -1),
            (1, 0),
            (1, 1),
        ];
        let mut future = Self::new(self.width, self.height);
        for h in 0..self.height {
            for w in 0..self.width {
                let alive = self.matrix[h * self.width + w];
                let neighbors = OFFSETS
                    .iter()
                    .filter(|offset| {
                        let height = (h as i64) + offset.0;
                        let width = (w as i64) + offset.1;
                        self.is_neighbor_alive(height, width)
                    })
                    .count();
                future.matrix[h * self.width + w] = if can_replicate(alive, neighbors) {
                    ALIVE
                } else {
                    DEAD
                };
            }
        }
        future
    }

    fn is_neighbor_alive(&self, height: i64, width: i64) -> bool {
        height >= 0
            && width >= 0
            && (height as usize) < self.height
            && (width as usize) < self.width
            && {
                let neighbor = self.matrix[height as usize * self.width + width as usize];
                neighbor.blue != 0 || neighbor.green != 0 || neighbor.red != 0
            }
    }

    /// Blits entire display
    fn render(&self, gop: &mut GraphicsOutput) -> Result {
        gop.blt(BufferToVideo {
            buffer: &self.matrix,
            src: BltRegion::Full,
            dest: (0, 0),
            dims: (self.width, self.height),
        })
    }
}

fn can_replicate(alive: BltPixel, neighbors: usize) -> bool {
    neighbors == 3 || (alive.blue != 0 || alive.green != 0 || alive.red != 0) && neighbors == 2
}

const BREEDER: &[(usize, usize)] = &[(415, 416), (416, 415), (416, 416), (417, 416), (417, 417)];

#[entry]
fn main() -> Status {
    uefi::helpers::init().unwrap();

    let mut life = Life::with_starter(800, 600, BREEDER);
    let gop_handle = boot::get_handle_for_protocol::<GraphicsOutput>().unwrap();
    let mut gop = boot::open_protocol_exclusive::<GraphicsOutput>(gop_handle).unwrap();

    for _ in repeat(()) {
        life = life.play();

        life.render(&mut gop).unwrap();
    }

    unreachable!()
}
